// SPDX-License-Identifier: Unlicense
pragma solidity ^0.8.24;

import "./Reputation.sol";
import "./Forum.sol";
import "../interfaces/IOnValidate.sol";
import "hardhat/console.sol";

struct ValidationPoolStake {
    uint id;
    bool inFavor;
    uint256 amount;
    address sender;
    bool fromMint;
}

struct ValidationPoolParams {
    uint duration;
    uint[2] quorum; // [ Numerator, Denominator ]
    uint[2] winRatio; // [ Numerator, Denominator ]
    uint bindingPercent;
    bool redistributeLosingStakes;
}

struct ValidationPool {
    uint id;
    uint postIndex;
    address sender;
    mapping(uint => ValidationPoolStake) stakes;
    uint stakeCount;
    ValidationPoolParams params;
    uint256 fee;
    uint endTime;
    bool resolved;
    bool outcome;
    bool callbackOnValidate;
    bytes callbackData;
}

contract ValidationPools is Reputation, Forum {
    mapping(uint => ValidationPool) public validationPools;
    uint public validationPoolCount;

    mapping(uint => address) public members;
    uint public memberCount;
    mapping(address => bool) public isMember;

    uint constant minDuration = 1; // 1 second
    uint constant maxDuration = 365_000_000 days; // 1 million years
    uint[2] minQuorum = [1, 10];

    event ValidationPoolInitiated(uint poolIndex);
    event ValidationPoolResolved(
        uint poolIndex,
        bool votePasses,
        bool quorumMet
    );

    // TODO: Add forum parameters

    /// Internal function to register a stake for/against a validation pool
    function _stakeOnValidationPool(
        ValidationPool storage pool,
        address sender,
        uint256 amount,
        bool inFavor,
        bool fromMint
    ) internal {
        require(block.timestamp <= pool.endTime, "Pool end time has passed");
        //_update(sender, address(this), amount);
        uint stakeIndex = pool.stakeCount++;
        ValidationPoolStake storage s = pool.stakes[stakeIndex];
        s.sender = sender;
        s.inFavor = inFavor;
        s.amount = amount;
        s.id = stakeIndex;
        s.fromMint = fromMint;
    }

    /// Accept reputation stakes toward a validation pool
    function stakeOnValidationPool(
        uint poolIndex,
        uint256 amount,
        bool inFavor
    ) public {
        ValidationPool storage pool = validationPools[poolIndex];
        _stakeOnValidationPool(pool, msg.sender, amount, inFavor, false);
    }

    /// Accept reputation stakes toward a validation pool
    function delegatedStakeOnValidationPool(
        uint poolIndex,
        address owner,
        uint256 amount,
        bool inFavor
    ) public {
        ValidationPool storage pool = validationPools[poolIndex];
        _spendAllowance(owner, msg.sender, amount);
        _stakeOnValidationPool(pool, owner, amount, inFavor, false);
    }

    /// Accept fee to initiate a validation pool
    /// TODO: Handle multiple authors
    function initiateValidationPool(
        uint postIndex,
        uint duration,
        uint[2] calldata quorum, // [Numerator, Denominator]
        uint[2] calldata winRatio, // [Numerator, Denominator]
        uint bindingPercent,
        bool redistributeLosingStakes,
        bool callbackOnValidate,
        bytes calldata callbackData
    ) external payable returns (uint poolIndex) {
        require(msg.value > 0, "Fee is required to initiate validation pool");
        require(duration >= minDuration, "Duration is too short");
        require(duration <= maxDuration, "Duration is too long");
        require(
            minQuorum[1] * quorum[0] >= minQuorum[0] * quorum[1],
            "Quorum is below minimum"
        );
        require(quorum[0] <= quorum[1], "Quorum is greater than one");
        require(winRatio[0] <= winRatio[1], "Win ratio is greater than one");
        require(bindingPercent <= 100, "Binding percent must be <= 100");
        Post storage post = posts[postIndex];
        require(post.author != address(0), "Target post not found");
        poolIndex = validationPoolCount++;
        ValidationPool storage pool = validationPools[poolIndex];
        pool.sender = msg.sender;
        pool.postIndex = postIndex;
        pool.fee = msg.value;
        pool.params.quorum = quorum;
        pool.params.winRatio = winRatio;
        pool.params.bindingPercent = bindingPercent;
        pool.params.redistributeLosingStakes = redistributeLosingStakes;
        pool.params.duration = duration;
        pool.endTime = block.timestamp + duration;
        pool.id = poolIndex;
        pool.callbackOnValidate = callbackOnValidate;
        pool.callbackData = callbackData;
        // We use our privilege as the DAO contract to mint reputation in proportion with the fee.
        // Here we assume a minting ratio of 1
        // TODO: Make minting ratio an adjustable parameter
        _mint(post.author, msg.value);
        // Here we assume a stakeForAuthor ratio of 0.5
        // TODO: Make stakeForAuthor an adjustable parameter
        _stakeOnValidationPool(pool, post.author, msg.value / 2, true, true);
        _stakeOnValidationPool(pool, post.author, msg.value / 2, false, true);
        emit ValidationPoolInitiated(poolIndex);
    }

    /// Evaluate outcome of a validation pool
    function evaluateOutcome(uint poolIndex) public returns (bool votePasses) {
        ValidationPool storage pool = validationPools[poolIndex];
        Post storage post = posts[pool.postIndex];
        require(pool.resolved == false, "Pool is already resolved");
        uint256 stakedFor;
        uint256 stakedAgainst;
        ValidationPoolStake storage s;
        for (uint i = 0; i < pool.stakeCount; i++) {
            s = pool.stakes[i];
            if (s.inFavor) {
                stakedFor += s.amount;
            } else {
                stakedAgainst += s.amount;
            }
        }
        // Special case for early evaluation if dao.totalSupply has been staked
        require(
            block.timestamp > pool.endTime ||
                stakedFor + stakedAgainst == totalSupply(),
            "Pool end time has not yet arrived"
        );
        // Check that quorum is met
        if (
            pool.params.quorum[1] * (stakedFor + stakedAgainst) <=
            totalSupply() * pool.params.quorum[0]
        ) {
            // Refund fee
            // TODO: this could be made available for the sender to withdraw
            // payable(pool.sender).transfer(pool.fee);
            pool.resolved = true;
            emit ValidationPoolResolved(poolIndex, false, false);
            // Callback if requested
            if (pool.callbackOnValidate) {
                IOnValidate(pool.sender).onValidate(
                    votePasses,
                    false,
                    stakedFor,
                    stakedAgainst,
                    pool.callbackData
                );
            }

            return false;
        }
        // A tie is resolved in favor of the validation pool.
        // This is especially important so that the DAO's first pool can pass,
        // when no reputation has yet been minted.

        votePasses =
            stakedFor * pool.params.winRatio[1] >=
            (stakedFor + stakedAgainst) * pool.params.winRatio[0];
        if (votePasses && !isMember[post.author]) {
            members[memberCount++] = post.author;
            isMember[post.author] = true;
        }
        pool.resolved = true;
        pool.outcome = votePasses;
        emit ValidationPoolResolved(poolIndex, votePasses, true);
        // Value of losing stakes should be distributed among winners, in proportion to their stakes
        uint256 amountFromWinners = votePasses ? stakedFor : stakedAgainst;
        uint256 amountFromLosers = votePasses ? stakedAgainst : stakedFor;
        // Only bindingPercent % should be redistributed
        // Stake senders should get (100-bindingPercent) % back
        // We have allowances for each stake. Time to collect from the losing stakes.
        uint totalRewards;
        uint totalAllocated;
        for (uint i = 0; i < pool.stakeCount; i++) {
            s = pool.stakes[i];
            uint bindingPercent = s.fromMint ? 100 : pool.params.bindingPercent;
            bool redistributeLosingStakes = s.fromMint ||
                pool.params.redistributeLosingStakes;
            if (votePasses != s.inFavor) {
                // Losing stake
                // If this stake is from the minted fee, don't burn it
                uint amount = (s.amount * bindingPercent) / 100;
                if (redistributeLosingStakes) {
                    _update(s.sender, address(this), amount);
                    totalRewards += amount;
                } else {
                    _burn(s.sender, amount);
                }
            }
        }
        // Issue rewards to the winners
        for (uint i = 0; i < pool.stakeCount; i++) {
            s = pool.stakes[i];
            uint bindingPercent = s.fromMint ? 100 : pool.params.bindingPercent;
            bool redistributeLosingStakes = s.fromMint ||
                pool.params.redistributeLosingStakes;
            if (redistributeLosingStakes && votePasses == s.inFavor) {
                // Winning stake
                // If this stake is from the minted fee, always redistribute it to the winners
                uint reward = (((amountFromLosers * s.amount) /
                    amountFromWinners) * bindingPercent) / 100;
                totalAllocated += reward;
                _update(address(this), s.sender, reward);
            }
        }
        // Due to rounding, some reward may be left over. Let's give it to the author.
        uint remainder = totalRewards - totalAllocated;
        if (remainder > 0) {
            _update(address(this), post.author, remainder);
        }

        // Distribute fee proportionately among all reputation holders
        for (uint i = 0; i < memberCount; i++) {
            address member = members[i];
            uint256 share = (pool.fee * balanceOf(member)) / totalSupply();
            // TODO: For efficiency this could be modified to hold the funds for recipients to withdraw
            payable(member).transfer(share);
        }

        // Callback if requested
        if (pool.callbackOnValidate) {
            IOnValidate(pool.sender).onValidate(
                votePasses,
                true,
                stakedFor,
                stakedAgainst,
                pool.callbackData
            );
        }
    }
}
